from enum import Enum
from abc import ABC, abstractmethod
from collections import namedtuple

import xarray as xr
import pandas as pd
import numpy as np

from toargridding.metadata import (
    Variables,
    Coordinates,
    get_global_attributes,
    AnalysisRequestResult,
    Metadata,
)
from toargridding.variables import Variable, Coordinate

GridType = Enum("GridType", ["regular"])


class GridDefinition(ABC):
    cell_index_name = "cell_index"

    def __init__(self):
        self.fill_value = np.nan  # TO(pd.DataFrame)) -> xr.Dataset:
        pass  # TODO make CF-compliant

    @staticmethod
    def construct(grid_type: GridType, **kwargs):
        match (grid_type):
            case GridType.regular:
                return RegularGrid(**kwargs)
            case other:
                raise Exception("invalid grid type")

    @property
    @abstractmethod
    def description(self):
        pass

    @staticmethod
    def from_netcdf():
        pass

    @abstractmethod
    def as_xarray(
        timeseries: dict[str, pd.DataFrame], metadata: pd.DataFrame
    ) -> dict[str, xr.Dataset]:
        pass


class RegularGrid(GridDefinition):
    Coord = namedtuple("Coord", ["lat", "lon"])

    def __init__(self, lat_resolution, lon_resolution):
        super().__init__()
        # TODO make sure only sensible resolutions

        self.lat = Coordinate.from_resolution(
            Coordinates.latitude, lat_resolution, min=-90, max=90
        )
        self.lon = Coordinate.from_resolution(
            Coordinates.longitude, lon_resolution, min=-180, max=180
        )
        spatial_shape = (self.lon.size, self.lat.size)
        spatial_size = self.lon.size * self.lat.size
        self.dims = [
            Coordinates.time.name,
            Coordinates.latitude.name,
            Coordinates.longitude.name,
        ]

        self._as_xy_index = np.dstack(
            np.meshgrid(range(self.lat.size), range(self.lon.size))
        ).reshape(-1, 2)
        self._as_i_index = np.arange(spatial_size).reshape(spatial_shape).T

    @property
    def description(self):
        return f"regular global grid with lat/lon resolutions ({self.lat.step}, {self.lon.step})"

    def as_xarray(self, data: AnalysisRequestResult) -> xr.Dataset:
        data_grouped_by_cell = self.group_data_by_cell(
            data.stations_data, data.stations_coords
        )
        cell_statistics = self.get_cell_statistics(data_grouped_by_cell)
        dataset = self.create_dataset(cell_statistics, data.metadata)

        return dataset

    def group_data_by_cell(self, data: pd.DataFrame, coords: pd.DataFrame):
        cell_indices = self.as_cell_index(coords)

        # will convert cell_indices to float as some nans ar present
        data_with_indices = data.join(
            cell_indices.to_frame(GridDefinition.cell_index_name), how="outer"
        )

        return data_with_indices.groupby(GridDefinition.cell_index_name)

    def get_cell_statistics(self, groups) -> dict[Variables, pd.DataFrame]:
        stats = {
            Variables.mean: groups.mean(),
            Variables.std: groups.std(),
            Variables.n: groups.count(),
        }

        return stats

    def create_dataset(self, cell_statistics: pd.DataFrame, metadata: Metadata):
        time = Coordinate.from_data(
            metadata.time.as_cf_index(),
            Coordinates.time,
            metadata,
            step=metadata.time.sampling,
        )

        gridded_ds = self.get_empty_grid(time, metadata)
        for variable, aggregated_data in cell_statistics.items():
            data_array_dict = self.get_data_array_dict(
                time, aggregated_data, variable, metadata
            )
            gridded_ds = gridded_ds.assign(data_array_dict)

        return gridded_ds

    def get_data_array_dict(self, time, aggregated_data, variable, metadata):
        gridded_data = self.create_gridded_data(time, aggregated_data)
        gridded_variable = Variable.from_data(gridded_data, variable, metadata)
        return {variable.name: gridded_variable.as_data_array(self.dims)}

    def create_gridded_data(self, time, grouped_timeseries):
        values = np.empty((time.size, self.lat.size, self.lon.size))
        values[...] = self.fill_value

        index = self._as_xy_index[grouped_timeseries.index.astype(int)]
        values[:, index.T[0], index.T[1]] = grouped_timeseries.values.reshape(
            -1, time.size
        ).T

        return values

    def as_cell_index(self, coords):
        id_x = self.coord_to_index(coords[self.lat.name], self.lat.min, self.lat.step)
        id_y = self.coord_to_index(coords[self.lon.name], self.lon.min, self.lon.step)

        id_i = self._as_i_index[id_x, id_y]

        return pd.Series(id_i, index=id_x.index)

    def coord_to_index(self, coord, x0_axis, d_axis):
        return (np.ceil((coord / d_axis) - 0.5) - x0_axis / d_axis).astype(int)

    def get_empty_grid(
        self, time: Variable, metadata: pd.DataFrame
    ) -> xr.Dataset:  # TODO make CF-compliant => docs
        coords = {
            Variables.time.name: time.as_data_array(),
            Variables.latitude.name: self.lat.as_data_array(),
            Variables.longitude.name: self.lon.as_data_array(),
        }

        ds = xr.Dataset(coords=coords, attrs=get_global_attributes(metadata))

        return ds
